import re
import warnings
from pprint import pprint

import googleapps

class XLSXWriterFormat(object):
    def __init__(self, **kwargs):
        for arg in kwargs:
            self.set(arg, kwargrs[arg])
    
    @classmethod 
    def render_border_json(cls, style=None, color=None):
        if style and style not in border_styles:
            raise Exception("Unknown border style '" + str(style) + "'")
        
        return ({
            'style': style,
            'color': cls.color_to_google_color(color)
            })
    
    
    @classmethod
    def translate_border_style(cls, style=1):
        return(cls.render_border_json(**xlsxwriter_border_styles[style]))
            
    
    def cellformat_json(self):
        json = {}
        raise Exception()
        
        #--------------------------------------------------------
        # numberFormat
        if self.number_format: 
            if type(sef.number_format) is int: # this is a hex value
                raise Exception("Have not implemented Excel's default formats yet")
                #json['numberFormat'] = {'format': self.number_format}
            else:
                json['numberFormat'] = {'pattern': self.number_format}
        
        #--------------------------------------------------------
        # backgroundColor    
        if self.background_color:
            if self.color_re(self.background_color):
                json['backgroundColor'] = self.render_colors(*self.bg_color)
            
        #--------------------------------------------------------
        # borders
        
        borders = {}
        if self.border_color:
            border_color = self.color_to_google_color(self.border_color)
        else: 
            # is this necessary? What happens if there is a border but no color?
            # is default black?
            border_color = self.color_to_google_color('black')

        # handle general/default border
        if self.border:
            # all borders
            border_def = {'style': self.translate_border_style(self.border),
                          'color': border_color}
            borders = {
                'top': border_def,
                'bottom': border_def,
                'left': border_def,
                'right': border_def,
            }
        
        # handle specific border data
        # Not sure if the color is/should be dependent on the style
        for border_type in ('top', 'bottom', 'left', 'right'):
            if getattr(self, border_type):
                borders.setdefault(border_type, {})
                borders[border_type]['style'] = self.translate_border_style(getattr(self, border_type))

                side_color = getattr(self, border_type_color)
                if side_color is None:
                    side_color = border_color    
                borders[border_type]['color'] = self.color_to_google_color(side_color)
                
                
        if borders:
            json['borders'] = borders            
            
        #--------------------------------------------------------
        # padding 
        # TODO 
         
         
        #--------------------------------------------------------
        # Alignment 
        if self.center_across or self.text_justlast:
            # these don't have google sheets equivalents, but we assume these are center-like
            # but if alignment is specifically specified, it will override this
            json["horizontalAlignment"] = 'center'
        
        if self.align:
            # in xlswriter, align can handle both horizontal and vertical
            if self.align in self.alignment_map:
                json["horizontalAlignment"] = self.alignment_map[self.align]
            else:
                json["verticalAlignment"] = self.vertical_alignment_map[self.align]
            
        if self.valign:
            # this can only be vertical - there is no setter in xlsxwriter for this
            json["verticalAlignment"] = self.vertical_alignment_map[self.vertical_align]

        if self.text_wrap is not None:
            if type(self.text_wrap) is bool:
                # this supports xlsxwriter, which is a boolean
                if self.text_wrap:
                    json["wrapStrategy"] = 'WRAP'
                else:
                    json["wrapStrategy"] = 'OVERFLOW_CELL'
            else:
                # allows full support of the google lexicon 
                json["wrapStrategy"] = self.text_wrap
                
        if self.rotation:
            json["textRotation"] = { 'angle': self.rotation }
            # also available is 'vertical': Boolean
            
        # this doesn't have an xlsxwriter equivalent        
        #    json["textDirection"] = self.
        #    json["hyperlinkDisplayType"] = self.


        text_format = {}
        if self.font_color:
            text_format['foregroundColor'] = self.color_to_google_color(self.font_color)
            
        if self.font_name:
            text_format["fontFamily"] = self.font_name

        if self.font_size:
            text_format["fontSize"] = self.font_size

        if self.bold:
            text_format["bold"] = self.bold

        if self.italic:
            text_format["italic"] = self.italic

        if self.font_strikeout:
            text_format["strikethrough"] = self.font_strikeout

        if self.underline:
            text_format["underline"] = self.underline


        if text_format:
            json["textFormat"] = text_format
        
        return( json )

        #--------------------------------------------------------
        
    wrap_strategies = ('OVERFLOW_CELL', 'LEGACY_WRAP', 'WRAP', 'CLIP')
        
    alignment_map = {'left':'LEFT', 'right':'RIGHT', 'center':'CENTER', None:'HORIZONTAL_ALIGN_UNSPECIFIED',
                     'justify':'center', 'center_across':'center'} # < these don't exist in google
        
    vertical_alignment_map = {'top':'TOP', 'vcenter': 'MIDDLE', 'bottom': 'BOTTOM', None:'VERTICAL_ALIGN_UNSPECIFIED',
                              'vjustify':'MIDDLE'} # < this doesn't exist in google
        
        #--------------------------------------------------------
                
    @property
    def font_name(self):
        return(self._font_name)

    @font_name.setter
    def set_font_name(self, value):
        self._font_name = value


    @property
    def font_size(self):
        return(self._font_size)

    @font_size.setter
    def set_font_size(self, value):
        self._font_size = value


    @property
    def font_color(self):
        return(self._font_color)

    @font_color.setter
    def set_font_color(self, value):
        self._font_color = value


    
    @property
    def bold(self):
        return(self._bold)

    @bold.setter
    def set_bold(self, value):
        self._bold = value


    @property
    def italic(self):
        return(self._italic)

    @italic.setter
    def set_italic(self, value):
        self._italic = value


    
    @property
    def underline(self):
        return(self._underline)

    @underline.setter
    def set_underline(self, value):
        self._underline = value


    @property
    def font_strikeout(self):
        return(self._font_strikeout)

    @font_strikeout.setter
    def set_font_strikeout(self, value):
        self._font_strikeout = value


    @property
    def font_script(self):
        return(self._font_script)

    @font_script.setter
    def set_font_script(self, value):
        self._font_script = value


    @property
    def num_format(self):
        return(self._num_format)

    @num_format.setter
    def set_num_format(self, value):
        self._num_format = value


    @property
    def locked(self):
        return(self._locked)

    @locked.setter
    def set_locked(self, value):
        self._locked = value


    @property
    def hidden(self):
        return(self._hidden)

    @hidden.setter
    def set_hidden(self, value):
        self._hidden = value


    @property
    def align(self):
        return(self._align)

    @align.setter
    def set_align(self, value):
        self._align = value


    @property
    def valign(self):
        return(self._valign)

    @valign.setter
    def set_valign(self, value):
        self._valign = value


    @property
    def rotation(self):
        return(self._rotation)

    @rotation.setter
    def set_rotation(self, value):
        self._rotation = value


    @property
    def text_wrap(self):
        return(self._text_wrap)

    @text_wrap.setter
    def set_text_wrap(self, value):
        self._text_wrap = value


    @property
    def text_justlast(self):
        return(self._text_justlast)

    @text_justlast.setter
    def set_text_justlast(self, value):
        self._text_justlast = value


    @property
    def center_across(self):
        return(self._center_across)

    @center_across.setter
    def set_center_across(self, value):
        self._center_across = value


    @property
    def indent(self):
        return(self._indent)

    @indent.setter
    def set_indent(self, value):
        self._indent = value


    @property
    def shrink(self):
        return(self._shrink)

    @shrink.setter
    def set_shrink(self, value):
        self._shrink = value


    @property
    def pattern(self):
        return(self._pattern)

    @pattern.setter
    def set_pattern(self, value):
        self._pattern = value


    @property
    def bg_color(self):
        return(self._bg_color)

    @bg_color.setter
    def set_bg_color(self, value):
        self._bg_color = value


    @property
    def fg_color(self):
        return(self._fg_color)

    @fg_color.setter
    def set_fg_color(self, value):
        self._fg_color = value


    @property
    def border(self):
        return(self._border)

    @border.setter
    def set_border(self, value):
        self._border = value


    @property
    def bottom(self):
        return(self._bottom)

    @bottom.setter
    def set_bottom(self, value):
        self._bottom = value


    @property
    def top(self):
        return(self._top)

    @top.setter
    def set_top(self, value):
        self._top = value


    @property
    def left(self):
        return(self._left)

    @left.setter
    def set_left(self, value):
        self._left = value


    @property
    def right(self):
        return(self._right)

    @right.setter
    def set_right(self, value):
        self._right = value


    @property
    def border_color(self):
        return(self._border_color)

    @border_color.setter
    def set_border_color(self, value):
        self._border_color = value


    @property
    def bottom_color(self):
        return(self._bottom_color)

    @bottom_color.setter
    def set_bottom_color(self, value):
        self._bottom_color = value


    @property
    def top_color(self):
        return(self._top_color)

    @top_color.setter
    def set_top_color(self, value):
        self._top_color = value


    @property
    def left_color(self):
        return(self._left_color)

    @left_color.setter
    def set_left_color(self, value):
        self._left_color = value


    @property
    def right_color(self):
        return(self._right_color)

    @right_color.setter
    def set_right_color(self, value):
        self._right_color = value

    @property
    def number_format_type(self):
        return(self._number_format_type)

    @number_format_type.setter
    def number_format_type(self, value):
        if type not in self.number_format_types:
            raise Exception("Unknown number format for Google Sheets: " + str(type))
        self._number_format_type = type
        
    @property
    def number_format_pattern(self):
        return(self._number_format_pattern)
    
    @number_format_pattern.setter
    def number_format_pattern(self, pattern):
        self._number_format_pattern = pattern
        
    @property
    def background_color(self):
        return(self._background_color)

    @background_color.setter
    def background_color(self, color): # color should be a tuple of (r,g,b [a])
        if type(color) is str:
            self._background_color = self.hex_to_color(color)
        else: self._background_color = color
            

    @property
    def borders(self):
        return(self._borders)

    @borders.setter
    def borders(self, data):
        self._borders = data

    @property
    def padding(self):
        return(self._padding)

    @padding.setter
    def padding(self, data):
        self._padding = data

    @property
    def alignment(self):
        return(self._alignment)

    @alignment.setter
    def alignment(self, data):
        self._alignment = data

    @property
    def vertical_alignment(self):
        return(self._vertical_alignment)

    @vertical_alignment.setter
    def vertical_alignment(self, data):
        self._vertical_alignment = data

    @property
    def text_wrap(self):
        return(self._text_wrap)

    @text_wrap.setter
    def text_wrap(self, data):
        self._text_wrap = data

    @property
    def text_direction(self):
        return(self._text_direction)

    @text_direction.setter
    def text_direction(self, data):
        self._text_direction = data

    @property
    def text_format(self):
        return(self._text_format)

    @text_format.setter
    def text_format(self, data):
        self._text_format = data

    @property
    def text_rotation(self):
        return(self._text_rotation)

    @text_rotation.setter
    def text_rotation(self, data):
        self._text_rotation = data

    @property
    def hyperlink(self):
        return(self._hyperlink)

    @hyperlink.setter
    def hyperlink(self, data):
        self._hyperlink = data




# Copyright (c) 2019-2021 Kevin Crouse
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# @license: http://www.apache.org/licenses/LICENSE-2.0
# @author: Kevin Crouse (krcrouse@gmail.com)

import os,sys,re
import re
import warnings
from pprint import pprint
import googleapps
import mimetypes

import email.encoders
import email.mime.text
import email.mime.multipart
from email.mime.text import MIMEText
from  email.mime.multipart import MIMEMultipart

from bs4 import BeautifulSoup
import base64

import time
import datetime
import googleapiclient
import io

class GMail(googleapps.ServiceClient):
    default_api = 'gmail.send'
    default_api_version = 'v1'

    def __init__(self, gmail_user):
        super().__init__()
        self.gmail_user = gmail_user


    def send_message(self, message=None, **kwargs):
        req = self._build_send_request(**kwargs)
        return(self.make_request(req))


    def add_message(self, **kwargs):
        req = self._build_send_request(**kwargs)
        return(self.add_request(req))


    def _build_send_request(self, message=None, callback=None, **kwargs):
        if not message:
            message = self.create_message(**kwargs)
        else:
            if len(args) or len(kwargs):
                raise Exception("Invalid arguments : " + str(args) + " | " + str(kwargs))

        request = ServiceRequest(resource=['users', 'messages'],
                         function='send',
                         callback=callback,
                         parameters={'userId': self.gmail_user, 'body': message})
        return(request)


    def create_message(self, to, subject, message, reply_to=None, sender=None, cc=None, bcc=None, markdown=False, html=None, text_message=None, attachments=None):
        """
            message {str}: A general message, which is processed as follows:
                If markdown is true, it is assumed to be a markdown (or markdown-html hybrid message) and it is converted to html
                If html is not None, it indicates whether message is already in html or in text.
                If html is None, message will be conservatively searched for a set of html-tag looking things, which will make the determination
        """

        msg = email.mime.multipart.MIMEMultipart()
        msg['Subject'] = subject

        if markdown:
            message = self.markdown_to_html(message)
        else:
            if html is None:
                # -- search for common tags
                if re.search(r'\<\s*(br|p|i|b|li|ul|ol|strong)\s*\/?\s*\>', message, re.I):
                    html = True
                else:
                    html = False

            if not html:
                text_message = message
                message = self.htmlify_text(message)

        if not text_message:
            text_message = self.textify_html(message)

        #print("---- message ----", message)
        #print("\n\n\n---- text message ----", text_message)
        msg.attach(email.mime.text.MIMEText(message, 'html'))
        #msg.attach(email.mime.text.MIMEText(text_message, 'plain'))

        if attachments:
            if type(attachments) not in (list, tuple):
                attachments = [ attachments ]
            for attachment in attachments:
                attachpayload = self.prepare_attachment(attachment)
                msg.attach(attachpayload)

        if sender:
            msg['From'] = sender
        if to:
            if type(to) in (list, tuple):
                msg['To'] = ','.join(to)
            else:
                msg['To'] = to

        if cc:
            if type(cc) in (list, tuple):
                msg['CC'] = ','.join(cc)
            else:
                msg['CC'] = cc

        if bcc:
            if type(bcc) in (list, tuple):
                msg['BCC'] = ','.join(bcc)
            else:
                msg['BCC'] = bcc

        if reply_to:
            if type(reply_to) in (list, tuple):
                msg['Reply-To'] = ','.join(reply_to)
            else:
                msg['Reply-To'] = reply_to

        raw = base64.urlsafe_b64encode(msg.as_bytes())
        raw = raw.decode()
        body = {'raw': raw}
        return(body)
        #return(email.encoders.encode_base64(msg))


    @classmethod
    def prepare_message(self, message, markdown=False, html=None, text_message=None):
        """
            message {str}: A general message, which is processed as follows:
                If markdown is true, it is assumed to be a markdown (or markdown-html hybrid message) and it is converted to html
                If html is not None, it indicates whether message is already in html or in text.
                If html is None, message will be conservatively searched for a set of html-tag looking things, which will make the determination
        """
        if markdown:
            message = self.markdown_to_html(message)
        else:
            if html is None:
                # -- search for common tags
                if re.search(r'\<\s*(br|p|i|b|li|ul|ol|strong)\s*\/?\s*\>', message, re.I):
                    html = True
                else:
                    html = False

            if not html:
                text_message = message
                message = self.htmlify_text(message)

        return(message)

    def prepare_attachment(self, filepath):

        my_mimetype, encoding = mimetypes.guess_type(filepath)

        # If the extension is not recognized it will return: (None, None)
        # If it's an .mp3, it will return: (audio/mp3, None) (None is for the encoding)
        #for unrecognized extension it set my_mimetypes to  'application/octet-stream' (so it won't return None again).
        if my_mimetype is None or encoding is not None:
            my_mimetype = 'application/octet-stream'


        main_type, sub_type = my_mimetype.split('/', 1)# split only at the first '/'
        # if my_mimetype is audio/mp3: main_type=audio sub_type=mp3

        #-----3.2  creating the attachement
            #you don't really "attach" the file but you attach a variable that contains the "binary content" of the file you want to attach

            #option 1: use MIMEBase for all my_mimetype (cf below)  - this is the easiest one to understand
            #option 2: use the specific MIME (ex for .mp3 = MIMEAudio)   - it's a shorcut version of MIMEBase

        #this part is used to tell how the file should be read and stored (r, or rb, etc.)
        if main_type == 'text':
            print("text")
            temp = open(filepath, 'r')  # 'rb' will send this error: 'bytes' object has no attribute 'encode'
            attachment = MIMEText(temp.read(), _subtype=sub_type)
            temp.close()

        elif main_type == 'image':
            print("image")
            temp = open(filepath, 'rb')
            attachment = email.mime.image.MIMEImage(temp.read(), _subtype=sub_type)
            temp.close()

        elif main_type == 'audio':
            print("audio")
            temp = open(filepath, 'rb')
            attachment = email.mime.audio.MIMEAudio(temp.read(), _subtype=sub_type)
            temp.close()

        elif main_type == 'application' and sub_type == 'pdf':
            from email.mime.application import MIMEApplication
            temp = open(filepath, 'rb')
            attachment = MIMEApplication(temp.read(), _subtype=sub_type)
            temp.close()

        else:
            attachment = email.mime.base.MIMEBase(main_type, sub_type)
            temp = open(filepath, 'rb')
            attachment.set_payload(temp.read())
            temp.close()

        #-----3.3 encode the attachment, add a header and attach it to the message
        email.encoders.encode_base64(attachment)  #https://docs.python.org/3/library/email-examples.html
        filename = os.path.basename(filepath)
        attachment.add_header('Content-Disposition', 'attachment', filename=filename)
        return(attachment)

    def prepare_attachment_GOOGLE(self, attachment):
          """ stolen from the google developer api """

          content_type, encoding = mimetypes.guess_type(attachment)

          if content_type is None or encoding is not None:
            content_type = 'application/octet-stream'
          main_type, sub_type = content_type.split('/', 1)
          if main_type == 'text':
            fp = open(attachment, 'rb')
            msg = MIMEText(fp.read(), _subtype=sub_type)
            fp.close()
          elif main_type == 'image':
            fp = open(attachment, 'rb')
            msg = email.mime.image.MIMEImage(fp.read(), _subtype=sub_type)
            fp.close()
          elif main_type == 'audio':
            fp = open(attachment, 'rb')
            msg = email.mime.audio.MIMEAudio(fp.read(), _subtype=sub_type)
            fp.close()
          else:
            fp = open(attachment, 'rb')
            msg = email.mime.base.MIMEBase(main_type, sub_type)
            msg.set_payload(fp.read())
            fp.close()
          filename = os.path.basename(attachment)
          return(filename, msg)

    @classmethod
    def textify_html(cls, html):
        text_message = html
        # invert links
        text_message = re.sub(r'<a[^>]+href\s*=\s*[\'"]\s*([^"\']+?)\s*[\'"].*?>(.*?)<\s*/a\s*>',
                              lambda x: x.group(2) + " (" + x.group(1) + ")  ",
                              text_message)
        # now paragraphs
        text_message = re.sub(r'<p>', "\n\n", text_message)

        #de-tag
        soup = BeautifulSoup(text_message, 'lxml')
        text_message = soup.get_text()
        return(text_message)

    @classmethod
    def htmlify_text(cls, text):
        # change linebreaks
        html_message = text
        html_message = re.sub(r"\n", '<br />', html_message)
        return(html_message)

    @classmethod
    def markdown_to_html(cls, msg, pretty=True):
        result_message = []
        in_list = False
        lines = msg.split("\n")
        htmlre = re.compile('(https?:\/\/[^\s]+)', re.I)
        for iline, line in enumerate(lines):

            # handle markdown links: [text label](http://href)
            line = re.sub(r'\[([^\]\r\n]+)\]\s*\(((http|ftp|mailto)[^\)\r\n]+)\)', lambda m: '<a href="'+m.group(2)+'">' + m.group(1) + '</a>', line)

            # handle header formatting
            m = re.match(r'(#+)\s+(.*)', line)
            if m:
                head_level =str(len(m.group(1)))
                line = '<h' + head_level  + '>' + m.group(2) + '</h' + head_level + '>'

            # before we do these, let's sub out html urls because wonky things happen
            presubs = {}
            presub_iter = 1
            mhtml = htmlre.search(line)
            while mhtml:
                repl = f'@@temp:{presub_iter}@@'
                line = line.replace(mhtml.group(1), repl)
                presubs[repl] = mhtml.group(1)
                presub_iter += 1
                mhtml = htmlre.search(line)

            # italics and bold
            line = re.sub(r'\*\*([^\*]+)\*\*', lambda m: '<b>' + m.group(1) + '</b>', line)
            line = re.sub(r'\_\_([^\_<>]+)\_\_', lambda m: '<b>' + m.group(1) + '</b>', line)
            #line = re.sub(r'\_([^\_<>]+)\_', lambda m: '<i>' + m.group(1) + '</i>', line)
            line = re.sub(r'\*([^\*]+)\*', lambda m: '<i>' + m.group(1) + '</i>', line)

            #
            # Now we can replace everythig in the line
            #
            for key, value in presubs.items():
                line = line.replace(key, value)

            # handle lists
            if re.match(r'\d+\.\s+', line):
                m = re.match(r'\d+\.\s+(.*)', line)
                if not in_list:
                    result_message.append('<ol>')
                    in_list = 'ol'
                result_message.append('<li>' + m.group(1) + '</li>')
            elif re.match(r'[\*\-\+]\s+', line):
                m = re.match(r'[\*\-\+]\s+(.*)', line)
                if not in_list:
                    result_message.append('<ul>')
                    in_list = 'ul'
                result_message.append('<li>' + m.group(1) + '</li>')
            elif re.match(r'\s*$', line):
            # This is a blank line.
            # if there is a blank line that is at the end of a list block, we just end the list
            # if there is a blank line that is not in a list context, we consider it an actual request for linebreak
                if in_list:
                    # any break ends the list
                    result_message.append('</'+in_list+'>')
                    in_list = False
                else:
                    result_message.append('<br />')
            elif iline == len(lines) - 1:
                # this is the last line of lines, so just add the text.
                result_message.append(line)
            else:
                # Not the last line!  So add the text and then a line break
                result_message.append(line + '<br />')

        # clean up any open tags
        if in_list:
            result_message.append('</'+in_list+'>')

        return(''.join(result_message))



# Copyright (c) 2019-2021 Kevin Crouse
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# @license: http://www.apache.org/licenses/LICENSE-2.0
# @author: Kevin Crouse (krcrouse@gmail.com)

import googleapps
from .paragraphelement import ParagraphElement
import re
heading_re = re.compile(r'HEADING_(\d+)')


@googleapps.jsonbase
class BodyElement(googleapps.JsonBase):
    json_properties = ['start_index', 'end_index']

    @classmethod
    def factory(cls, json, parent):
        if 'paragraph' in json:
            return(ParagraphBody(json, parent))
        elif 'sectionBreak' in json:
            return(SectionBreakBody(json, parent))
        elif 'table' in json:
            return(TableBody(json, parent))
        elif 'tableOfContents' in json:
            return(TableOfContentsBody(json, parent))
        else:
            raise Exception("Cannot identify body element type")

class SectionBreakBody(BodyElement):
    body_type = 'section break'
    json_properties = BodyElement.json_properties + ['section_break']


class TableBody(BodyElement):
    body_type = 'table'
    json_properties = BodyElement.json_properties + ['table']

class TableOfContentsBody(BodyElement):
    body_type = 'table of contents'
    json_properties = BodyElement.json_properties + ['table_of_contents']

class ParagraphBody(BodyElement):
    body_type = 'paragraph'
    json_properties = BodyElement.json_properties + ['paragraph']

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._elements = None
        self._style = None

    @property
    def bullet(self):
        if 'bullet' in self.paragraph:
            return(self.paragraph['bullet'])

    @property
    def list(self):
        if not self.bullet:
            return
        l = self.parent.lists[self.bullet['listId']]
        if 'nestingLevel' in self.bullet:
            return(l.get_level(self.bullet['nestingLevel']))
        else:
            return(l.get_level(0))

    @property
    def list_type(self):
        l = self.list
        if not l:
            return
        return(l.list_type)

    @property
    def style(self):
        if not self._style:
            self._style = ParagraphStyle(self.paragraph['paragraphStyle'])
        return(self._style)

    @property
    def paragraph_type(self):
        if self.style.name == 'NORMAL_TEXT':
            return('text')
        elif heading_re.match(self.style.name):
            return('heading')
        elif self.style.name == 'TITLE':
            return('title')
        elif self.style.name == 'SUBTITLE':
            return('subtitle')

    @property
    def heading_level(self):
        m = heading_re.match(self.style.name)
        if not m:
            return
        return(int(m.group(1)))


    @property
    def elements(self):
        if not self._elements:
            self._elements = [ParagraphElement(elem, self) for elem in self.paragraph['elements']]
        return(self._elements)


class ParagraphStyle(googleapps.JsonBase):

    @property
    def name(self):
        return(self.json['namedStyleType'])



# Copyright (c) 2019-2021 Kevin Crouse
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# @license: http://www.apache.org/licenses/LICENSE-2.0
# @author: Kevin Crouse (krcrouse@gmail.com)


from pprint import pprint
import googleapps
from googleapps.serviceclient import CommitType
from .bodyelement import BodyElement
from .listelement import ListElement


@googleapps.serviceclient.googleappclient
class Document(googleapps.ServiceClient):
    default_api = 'docs'
    default_api_version = 'v1'
    api_ro_properties = ('document_id', 'revision_id', 'suggestions_view_mode', 'inline_objects')
    api_properties = ('title', )
    commit_policy=CommitType.defer_commit

    def __init__(self, *args, **kwargs):
        self._json = None
        self._body = None
        self._lists = None
        # can create a drive item as :
        # Document(id)

        # Document(name=TITLE) - create new file

        # TODO: Verify default behavior of DriveItem(single_param) is that single_param is the name. Not the ID.
        if len(args) == 1:
            kwargs['document_id'] = args[0]
        elif len(args) > 1:
            raise Exception("googleapps.drive objects do not provide support for more than one ordered argument")

        super().__init__(**kwargs)

    @property
    def id(self):
        return(self.document_id)

    def load(self):
        self.make_request('documents', 'get', documentId=self.id, callback=self.process_json)

    def refresh(self):
        # at present this is no different than load
        # but we keep it separate in case we want to abstract it further sometime in the future
        self.load()

    def process_json(self, djson):
        self._document_id  = djson['documentId']
        self._json = djson
        for param in self.api_properties:
            setattr(self, '_' + param, djson[googleapps.toCamel(param)])

    @property
    def json(self):
        if not self._json:
            self.load()
        return(self._json)

    @property
    def body(self):
        if not self._body:
            self._body = [BodyElement.factory(x, self) for x in self.json['body']['content']]
        return(self._body)

    @property
    def lists(self):
        if not self._lists:
            self._lists = {key:ListElement(key, val, self) for key, val in self.json['lists'].items()}
        return(self._lists)



import googleapps
from .driveitem import DriveItem
from .file import File
from .folder import Folder

from googleapps.serviceclient import CommitType
import googleapiclient

import os, os.path
import datetime
import pickle
import re
import shutil
import enum

import warnings
from pprint import pprint

def item_class(mime_type=None):
    if mime_type == MimeType.folder_mime_type():
        return(Folder)
    else:
        return(File)

def item_from_json(json):
    #
    # This can be expanded as we have more classes
    #


    #print("******* Item from json ******")
    #pprint(json)
    #print("*****************************")
    ItemType = item_class(json['mimeType'])
    item = ItemType(id=json['id'])
    item.process_request_json(json)
    return(item)



def create_folder(*args, **kwargs):
    kwargs['mime_type'] = MimeType.folder_mime_type()
    return(create(*args, **kwargs))


def create(name=None, in_folder=None, path=None, **body):

    if 'mime_type' in body:
        ItemClass = item_class(body['mime_type'])
    else:
        ItemClass = item_class()

    if name:
        body['name'] = name

    if in_folder:
        body['parents'] = [in_folder]

    if path:
        raise Exception("Haven't implemented path logic yet")

    item = ItemClass(**body)

    # only save if autocommit is on
    if item.commit_policy >= CommitType.auto_commit:
        item.save()

    return(item)


def upload(filepath, mime_type=None, convert=None, **parameters):
    obj = File(**parameters, mime_type=mime_type)
    obj.upload(filepath, convert=convert)
    return(obj)

def get(id, retry_count=3, search_all_drives=True, **extra_args):
    if extra_args:
        extra_args = {googleapps.toCamel(key): extra_args[key] for key in extra_args}
    else:
        extra_args = {}

    request = googleapps.serviceclient.ServiceRequest(api=DriveItem.default_api,
                                                      api_version=DriveItem.default_api_version,
                                                      resource='files',
                                                      function='get',
                                                      parameters={'fileId':id, 'fields':'*', 'supportsAllDrives':search_all_drives, **extra_args},
                                                      callback=None
                                                      )
    try:
        response = DriveItem.make_class_request( request, attempt_recover=False )
        return(item_from_json(response))
    except googleapiclient.errors.HttpError as e:
        if re.search('File not found', str(e)):
            # if the file is not found, we just return null
            return
        elif re.search('Resource not avail|Try again later', str(e)):

            if retry_count:
                # if we suspect this is a network or resource busy error, we just call get again
                return(get(id, retry_count-1))
            else:
                # other errors, pass along
                raise(e)
        else:
            print("Not sure what to do with this HttpError")
            raise(e)

def open(id):
    return(get(id))

def open_folder(id):
    folder = get(id)
    if folder is None:
        return

    if folder.mime_type != MimeType.folder_mime_type():
        raise Exception("Attempted to open folder '" + str(id) + "', but it is not a Folder! It is a " + folder.mime_type)

    return(folder)


def open_path(path, in_folder=None):
    raise Exception("Not implemented yet")

def exists(path, in_folder=None):
    raise Exception("Not implemented yet")


def find(*args, **kwargs):
    # shortcut to find exactly one item
    result = find_all(*args, **kwargs)
    if not result:
        return
    if len(result) == 1:
        return result[0]
    raise Exception("Parameters sent to find_file() matched multiple files: " + "; ".join(f.name + '( ' + f.id + ' )' for f in result))



def find_all(filename=None, file_type=None, in_folder=None, domains=None, spaces=None, order=None, all_drives=None, **queryfields):

    params = { }
    if domains:
        params['corpora'] = domains

    if all_drives is not None:
        #params['include_items_from_all_drives'] = all_drives
        params['supportsAllDrives'] = all_drives
        params['includeItemsFromAllDrives'] = all_drives

    if order:
        if 'order_by' in queryfields:
            raise Exception('Cannot specify both order_by and order')
        params['order_by'] = order

    if spaces:
        params['spaces'] = spaces

    if filename:
        if 'name' in queryfields:
            raise Exception("Cannot specify both name and filename. These are the same")
        queryfields['name'] = filename

    if file_type:
        if 'mime_type' in queryfields:
            raise Exception('Cannot specify both file_type and mime_type')
        queryfields['mime_type'] = MimeType.mime_type(file_type)

    if in_folder:
        if 'parents' in queryfields:
            raise Exception("Cannot specify in_folder and parents")
        if type(in_folder) is str:
            queryfields['parents'] = {'op': 'in', 'value': in_folder}
        else:
            queryfields['parents'] = {'op': 'in', 'value': in_folder.id}

    # by default, we do not serch in the trash
    if 'trashed' not in queryfields:
        queryfields['trashed'] = False

    query = DriveItem.generate_query_string(**queryfields)

    #print("------------------------------ FIND ALL -------------------------------------------")
    #print(query)
    #print("------------------------------ END FIND ALL -------------------------------------------")
    response = DriveItem.services.load_service('drive').files().list(
        fields='*',
        q=query,
        **params
    ).execute()

    result = []
    #--- if there are more files, loop over them !
    another_page = True
    while another_page:
        for f in response['files']:
            result.append(item_from_json(f))

        another_page = response.get('nextPageToken', None)
        if another_page:
            if googleapps.ServiceClient.debug:
                print("-- Fetching another page worth of files because it appears there are more to download")
            response = DriveItem.services.load_service('drive').files().list(
                fields='*',
                q=query,
                pageToken=another_page,
                **params
            ).execute()


    return(result)


class MimeType():

    core_mime_types = [
        ('file', 'application/vnd.google-apps.file'),
        ('folder', 'application/vnd.google-apps.folder'),
        ('unknown', 'application/vnd.google-apps.unknown'),
        ('bmp', 'image/bmp'),
        ('gif', 'image/gif'),
        ('jpeg', 'image/jpeg'),
        ('png', 'image/png'),
        ('svg', 'image/svg+xml'),
        ('pdf', 'application/pdf'),
        ('css', 'text/css'),
        ('csv', 'text/csv'),
        ('html', 'text/html'),
        ('javascript', 'application/javascript'),
        ('plain_text', 'text/plain'),
        ('rtf', 'application/rtf'),
        ('zip', 'application/zip'),
        ('google_app_script', 'application/vnd.google-apps.script'),
        ('google_drawing', 'application/vnd.google-apps.drawing'),
        ('google_docs', 'application/vnd.google-apps.document'),
        ('google_forms', 'application/vnd.google-apps.form'),
        ('google_sheets', 'application/vnd.google-apps.spreadsheet'),
        ('google_slides', 'application/vnd.google-apps.presentation'),
        ('opendocument_graphics', 'application/vnd.oasis.opendocument.graphics'),
        ('opendocument_presentation', 'application/vnd.oasis.opendocument.presentation'),
        ('opendocument_spreadsheet', 'application/vnd.oasis.opendocument.spreadsheet'),
        ('opendocument_text', 'application/vnd.oasis.opendocument.text'),
        ('microsoft_excel', 'application/vnd.openxmlformats-officedocument.spreadsheetml.sheet'),
        ('microsoft_excel_legacy', 'application/vnd.ms-excel'),
        ('microsoft_powerpoint', 'application/vnd.openxmlformats-officedocument.presentationml.presentation'),
        ('microsoft_powerpoint_legacy', 'application/vnd.ms-powerpoint'),
        ('microsoft_word', 'application/vnd.openxmlformats-officedocument.wordprocessingml.document'),
        ('microsoft_word_legacy', 'application/msword'),
    ]

    #
    extensions = [
        ('odg', 'application/vnd.oasis.opendocument.graphics'),
        ('odp', 'application/vnd.oasis.opendocument.presentation'),
        ('ods', 'application/vnd.oasis.opendocument.spreadsheet'),
        ('odt', 'application/vnd.oasis.opendocument.text'),
        ('xlsx', 'application/vnd.openxmlformats-officedocument.spreadsheetml.sheet'),
        ('xls', 'application/vnd.ms-excel'),
        ('pptx', 'application/vnd.openxmlformats-officedocument.presentationml.presentation'),
        ('ppt', 'application/vnd.ms-powerpoint'),
        ('docx', 'application/vnd.openxmlformats-officedocument.wordprocessingml.document'),
        ('doc', 'application/msword'),
        ('jpg', 'image/jpeg'),
        ('js', 'application/javascript'),
        ('txt', 'text/plain'),
        ('csv', 'text/csv'),
    ]

    # Some aliases for simplicity
    aliases = [
        ('google_document', 'google_docs'),
        ('google_spreadsheet', 'google_sheets'),
        ('google_presentation', 'google_slides'),
        ('libreoffice_graphics', 'opendocument_graphics'),
        ('libreoffice_document', 'opendocument_text'),
        ('libreoffice_presentation', 'opendocument_presentation'),
        ('libreoffice_spreadsheet', 'opendocument_spreadsheet'),
        ('openoffice_graphics', 'opendocument_graphics'),
        ('openoffice_document', 'opendocument_text'),
        ('openoffice_presentation', 'opendocument_presentation'),
        ('openoffice_spreadsheet', 'opendocument_spreadsheet'),
        ('ms_excel', 'microsoft_excel'),
        ('ms_excel_legacy', 'microsoft_excel_legacy'),
        ('ms_powerpoint', 'microsoft_powerpoint'),
        ('ms_powerpoint_legacy', 'microsoft_powerpoint_legacy'),
        ('ms_word', 'microsoft_word'),
        ('ms_word_legacy', 'microsoft_word_legacy'),
    ]

    # mime_types that can be translated to Google Apps Office Formats
    office_mime_types = {
        'application/rtf': 'application/vnd.google-apps.document',
        'application/vnd.openxmlformats-officedocument.wordprocessingml.document': 'application/vnd.google-apps.document',
        'application/vnd.oasis.opendocument.text': 'application/vnd.google-apps.document',
        'application/vnd.google-apps.document': 'application/vnd.google-apps.document',

        'application/vnd.google-apps.spreadsheet': 'application/vnd.google-apps.spreadsheet',
        'text/csv': 'application/vnd.google-apps.spreadsheet',
        'text/tab-separated-values': 'application/vnd.google-apps.spreadsheet',
        'application/vnd.openxmlformats-officedocument.spreadsheetml.sheet': 'application/vnd.google-apps.spreadsheet',
        'application/x-vnd.oasis.opendocument.spreadsheet': 'application/vnd.google-apps.spreadsheet',

        'application/vnd.google-apps.presentation': 'application/vnd.google-apps.presentation',
        'application/vnd.oasis.opendocument.presentation': 'application/vnd.google-apps.presentation',
        'application/vnd.openxmlformats-officedocument.presentationml.presentation': 'application/vnd.google-apps.presentation',
    }

    google_document_types = {
        'application/vnd.google-apps.document',
        'application/vnd.google-apps.drawing',
        'application/vnd.google-apps.form',
        'application/vnd.google-apps.presentation',
        'application/vnd.google-apps.spreadsheet',
    }

    office_conversion_types = {
        'document': 'application/vnd.google-apps.document',
        'spreadsheet': 'application/vnd.google-apps.spreadsheet',
        'presentation': 'application/vnd.google-apps.presentation',
    }

    office_types = {
        'application/rtf': 'document',
        'application/vnd.openxmlformats-officedocument.wordprocessingml.document': 'document',
        'application/vnd.oasis.opendocument.text': 'document',
        'application/vnd.google-apps.document': 'document',

        'application/vnd.google-apps.spreadsheet': 'spreadsheet',
        'text/csv': 'spreadsheet',
        'text/tab-separated-values': 'spreadsheet',
        'application/vnd.openxmlformats-officedocument.spreadsheetml.sheet': 'spreadsheet',
        'application/x-vnd.oasis.opendocument.spreadsheet': 'spreadsheet', #libreoffice/ods
        'application/vnd.oasis.opendocument.spreadsheet': 'spreadsheet', #libreoffice/ods
        'application/vnd.ms-excel': 'spreadsheet', # legacy msexcel

        'application/vnd.google-apps.presentation': 'presentation',
        'application/vnd.oasis.opendocument.presentation': 'presentation',
        'application/vnd.openxmlformats-officedocument.presentationml.presentation': 'presentation',
    }

    media_types = ( 'application/vnd.google-apps.audio',
                    'application/vnd.google-apps.video',
                    'application/vnd.google-apps.photo')


    _mime_types = {}
    _reverse_mime_types = {}
    _aliases = {}
    _reverse_aliases= {}
    _extensions = {}
    _reverse_extensions = {}

    for mt in core_mime_types:
        _mime_types[mt[0]] = mt[1]
        _reverse_mime_types[mt[1]] = mt[0]

    for al in aliases:
        _aliases[al[0]] = al[1]
        _reverse_aliases[al[1]] = al[0]

    for ext in extensions:
        _extensions[ext[0]] = ext[1]
        _reverse_extensions[ext[1]] = ext[0]

    @classmethod
    def folder_mime_type(cls):
        return(cls._mime_types['folder'])

    @classmethod
    def mime_type_from_filename(cls, filename):
        ext = re.search(r'\.([^\.]+)$')
        if ext:
            return(cls.extension_mime_type(ext.group(1)))
        return

    @classmethod
    def extension_mime_type(cls, ext):
        if ext.lower() in cls._extensions:
            return(cls._extensions[ext.lower()])
        return

    @classmethod
    def mime_type_extension(cls, mime_type):
        if mime_type.lower() in cls._reverse_extensions:
            return(cls._reverse_extensions[mime_type.lower()])
        return

    @classmethod
    def mime_type(cls, file_type):
        # see if the type provided is a direct link
        if file_type.lower() in cls._mime_types:
            return(cls._mime_types[file_type.lower()])
        # otherwise, see what the alias is
        if file_type.lower() in cls._aliases:
            return(cls._mime_types[cls._aliases[file_type.lower()]])

        # default is to just return the mime_type for "file"
        return(cls._mime_types['file'])

    @classmethod
    def file_type(cls, mime_type):
        # see if the type provided is a direct link
        if mime_type.lower() in cls._reverse_mime_types:
            return(cls._reverse_mime_types[mime_type.lower()])

        # default is just to return 'file'
        return('file')

    @classmethod
    def document_type(cls, mime_type=None, file_type=None, extension=None):
        if mime_type:
            mime_type = mime_type.lower()
        elif file_type:
            mime_type = cls.mime_type(file_type)
        elif extension:
            mime_type = cls.extension_mime_type(extension)
        if mime_type in cls.office_types:
            return(cls.office_types[mime_type])
        return

    @classmethod
    def google_document_type(cls, mime_type=None, file_type=None, extension=None):
        if mime_type:
            mime_type = mime_type.lower()
        elif file_type:
            mime_type = cls.mime_type(file_type)
        elif extension:
            mime_type = cls.extension_mime_type(extension)

        if re.search('google-apps', mime_type) and mime_type in cls.office_types:
            return(cls.office_types[mime_type])
        return


    @classmethod
    def convert_document_mime_type(cls, **kwargs):
        office_type = cls.document_type(**kwargs)
        if not office_type:
            return
        return(cls.office_conversion_types[office_type])

for mt in MimeType.core_mime_types:
    setattr(MimeType, mt[0], mt[1])

for alias in MimeType.aliases:
    setattr(MimeType, alias[0], getattr(MimeType,alias[1]))



import setuptools
with open("README.md", "r") as fh:
    long_description = fh.read()

with open("VERSION.txt", "r") as fh:
    version = fh.read().strip()

setuptools.setup(
    name='pygoogleapps',
    version=version,
    author="Kevin Crouse",
    author_email="krcrouse@gmail.com",
    description='A module to provide a stateful wrapper around the Google App system',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://gitlab.com/krcrouse/pygoogleapps",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3.7",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
        "Development Status :: 3 - Alpha"
    ],    
    install_requires=[
        'python-dateutil',
        'google.auth',
        'google-api-python-client',
        'google_auth_oauthlib',
    ],
)



# Copyright (c) 2019-2021 Kevin Crouse
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# @license: http://www.apache.org/licenses/LICENSE-2.0
# @author: Kevin Crouse (krcrouse@gmail.com)

from pprint import pprint
import googleapps
import datetime
from .event import *


@googleapps.serviceclient.googleappclient
class Calendar(googleapps.ServiceClient):
    default_api = 'calendar'
    default_api_version = 'v3'
    api_ro_properties = ('id', 'summary', 'description')
    #default_scope = 'https://www.googleapis.com/auth/calendar.readonly'
    default_scope = 'https://www.googleapis.com/auth/calendar'

    def __init__(self, *args, **kwargs):
        self._json = None
        # TODO: Verify default behavior of DriveItem(single_param) is that single_param is the name. Not the ID.
        if len(args) == 1:
            kwargs['id'] = args[0]
        elif len(args) > 1:
            raise Exception("googleapps.drive objects do not provide support for more than one ordered argument")

        super().__init__(**kwargs)
        self._events = None

    @property
    def calendar_id(self):
        return(self.json['id'])

    @property
    def json(self):
        if not self._json:
            self.load()
        return(self._json)

    def load(self):
        self.make_request('calendars', 'get', calendarId=self.id, callback=self.process_json)

    def process_json(self, jresult):
        self._id  = jresult['id']
        self._json = jresult
        for param in self.api_properties:
            setattr(self, '_' + param, jresult[googleapps.toCamel(param)])

    def events(self, min_date=None, max_date=None):
        if not min_date:
            min_date=datetime.datetime.now()
        if not self._events:
            eventjson = self.make_request('events', 'list', calendarId=self.id, timeMin=min_date)
            self._events = EventList(eventjson, self)
        return(self._events)

    def create_event(self, summary, start, end, **kwargs):
        e = Event.create(self, self.id, summary, start, end, **kwargs)
        self._events.append(e)


# Copyright (c) 2019-2021 Kevin Crouse
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# @license: http://www.apache.org/licenses/LICENSE-2.0
# @author: Kevin Crouse (krcrouse@gmail.com)

import googleapps
import re
import datetime

class JSONProperty():

    def __init__(self, property, readonly=False):
        self.property = property
        self.readonly = readonly
        self.attr = '_' + property

    def __get__(self, obj, cls=None):
        if not hasattr(obj, self.attr):
            return(None)
        return(getattr(obj, self.attr))

    def __set__(self, obj, value):
        return(setattr(obj, self.attr, value))

    def __delete__(self, obj):
       raise Exception("Not supporting json property setting yet")


class JSONObject():

    def __init__(self, property, object_class):
        self.property = property
        self.attr = '_' + property
        self.object_class = object_class

    def __get__(self, obj, cls=None):
        if not hasattr(obj, self.attr):
            return(None)
        return(getattr(obj, self.attr))

    def __set__(self, obj, value):
        # -- actually, this expects the value to be a json object that it passes to the class
        klass = self.object_class
        if type(value) in (list, tuple):
            result = []
            for item in value:
                result.append(klass(item, parent=self))
            return(setattr(obj, self.attr, result))
        else:
            setattr(obj, self.attr, klass(value, parent=self))

    def __delete__(self, obj):
       raise Exception("Not supporting json property setting yet")


def jsonbase(cls):
    # This is just shorthand for a specific interface for the classproperties
    for prop in cls.json_properties:
        setattr(cls, prop, JSONProperty(prop))

    for prop, objclass in cls.json_objects.items():
        setattr(cls, prop, JSONObject(prop, objclass))
    return(cls)

isodate = re.compile(r'(\d\d\d\d)-(\d\d)-(\d\d)')
class JsonBase():
    json_objects = {}
    json_properties = []

    def __init__(self, json, parent=None):
        self._parent = parent
        self.set_defaults()
        if json is None:
            self._json = None
        else:
            self.process_json(json)

    def set_defaults(self):
        return

    @property
    def json(self):
        return(self._json)

    @property
    def parent(self):
        return(self._parent)

    def process_json(self, jresult):
        self._json = jresult
        for param in self.json_properties + list(self.json_objects):
            camelParam = googleapps.toCamel(param)
            if camelParam in jresult:
                val = jresult[camelParam]
                if type(val) is str:
                    datepieces = googleapps.from_rfctime(val)
                    if datepieces:
                        val = datepieces
                    else:
                        dm = isodate.fullmatch(val)
                        if dm:
                            val = datetime.date(int(dm.group(1)), int(dm.group(2)), int(dm.group(3)))

                setattr(self, param, val)



# Copyright (c) 2019-2021 Kevin Crouse
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# @license: http://www.apache.org/licenses/LICENSE-2.0
# @author: Kevin Crouse (krcrouse@gmail.com)

import os, os.path
import datetime
import re
import shutil

import warnings
from pprint import pprint

import pickle
import google.auth, google.oauth2
import google_auth_oauthlib
import googleapiclient.discovery # to install this : pip install --upgrade google-api-python-client

#-------------- Notes on Permission Scope
# Permission Scopes are the Google-designed access scopes for all of the
# services provided. The generic GoogleApps.Service interface allows for all scopes,
# but the valuable elements of the suite will be missing. This allows you to create
# your own supported interfaces even if you don't intend to share.  But please share
# if it's any good.

# At present, the full list of Google Permission Scopes are listed here:
# https://developers.google.com/identity/protocols/googlescopes

# As specific interfaces are designed and added to this suite of Python GoogleApps package,
# they should be added to the directory below for auto-resolution

# We process scopes based on single intuitive keywords, but Google has a number of identifiers.
# This is the location where we map them. Users can also specify these independently

#  1. the Google API Name, which is a shorthand, clear descriptor
#  2. The API version
#  3. The Scope

DEFAULT_SCOPES = {
    'gmail': ['gmail', 'v1', 'gmail.send'],
    'sheets': ['sheets', 'v4', 'spreadsheets'],
    'sheets.readonly': ['sheets', 'v4', 'spreadsheets.readonly'],
    'drive': ['drive', 'v3', 'drive'],
    'drive.metadata': ['drive', 'v3', 'drive.metadata'],
    'drive.readonly': ['drive', 'v3', 'drive.readonly'],
    'docs': ['docs', 'v1', 'documents'],
}

class ServiceManager(object):
    '''
    The ServiceManager is intended to encapsulate all of the Google services for a user. The SM object must be tied to the registered OAUTH secret file, which the user must download from the Google Developer Console (https://console.developers.google.com/apis/credentials). The filename from Google will be "client_secret_{id}.json", but if you want to keep your scripts simple, the ServiceManager default looks for "google_client.json" in the "credentials" directory. Of course, these can be overriden and you can manually point to a client secret file.

    Notes:
        Services are cached by default. So, attempt to access the Sheets service from two different parts of an application will refer to the same service and service manager. This is done so you don't have to keep track of service manager objects in modules that are designed to be independent of each other, and, in most cases, this is the correct model because it ends up functioning like a singleton as only one user is authenticating to Google for a given execution. However, since there are potential needs to reference multiple paths, it can be explicitly instantiated to two different paths, in which you will need to balance multiple users, each connected to a separate service managers.
    '''
    api_key=None
    secret_file = 'google_client.json'
    credential_directory='credentials'
    oauth_args = []

    loaded_services = {}

    @classmethod
    def clear_class_services(cls):
        """ Uncache all services for this ServiceManager. """
        cls.loaded_services = {}

    def detach_services(self):
        """
        This will transfer the services from class variables to instance variables, for use with multiple ServiceManagers that point to different credentials. This has two purposes:
          1. It creates local services when the end developer doesn't want services pooled at the class level
          2. If there are already services previously loaded at the instance level, it clears them.
        """
        self.loaded_services = self.__class__.loaded_services.copy()

    def __init__(self,
                 services=None,
                 api_key=None,
                 secret_file=None,
                 credential_directory=None,
                 ):
        """
        Args:
            services (list, optional): A list of services to load up initially.
            api_key (str, optional): Uses the proivded developer API Key instead of OAUTH2 for authentication.
            secret_file (filename): The filename of the Google Developer OAUTH Credential File. By default, this is set to google_client.json. If no file exists, an exception is thrown on first attempt to authenticate.
            credential_directory (path): The path to the credential directory where the OATH credential file is and where service authorization tokens should be stored. By default, this is the credentials directory from wherever the program is being run. There are no built-in options to have a default  credential location because any guest would potentially be able to access all the user's Google data.
        """

        if api_key:
            self.api_key=api_key
        if credential_directory:
            self.credential_directory = credential_directory
        if secret_file:
            self.secret_file = secret_file
        if services:
            self.load_services(services)

    def load_services(self, service_keys):
        """ A proxy function to load a list of services instead of a single service. """
        for service in service_keys:
            self.load_service(service)

    def get_service(self, service_keys, load_if_invalid=False):
        """ Fetches *one* of the services requested, giving priority first to the existing cache or by loading them new. In most cases, this is handled by api-specific module.

        This is used to minimize new manual authorizations (or authorizations that exceed the necessary permissions) by fetching one of multiple permission scopes when a client API call can be called by multiple permissions scopes. For example, reading a description for an API object often can be done within a metadata scope, readonly scope, and full access scope. If no authorizations have been provided, we should load a service that only asks for metadata access, as it's the least intrusive. However, if the user has *already* authorized full access, there's no reason to launch a separate authentication window for the metadata scope, as it is entirely encompassed in full access.

        In the case a new authentication is required, the first service in the service_keys array is chosen as the scope. The user can change the order of the keys then to minimize authentications (e.g. listing full access before readonly access if their application will later be requiring full access).
        Args:
            service_keys (str|list): The list of valid service keys to search over.
            load_if_invalid (bool): Automatically load the first service in the service_keys list if no service currently exists. Default is False.
        Returns:
            A single, loaded service of the provided service_keys, or None if no services are found and load_if_invalid is False.
        """

        if type(service_keys) is not list:
            service_keys = [ service_keys ]

        for service_key in service_keys:
            if service_key in self.loaded_services:
                return(self.loaded_services[service_key])

        if load_if_invalid:
            return(self.load_service(service_keys[0]))
        else:
            return

    def load_service(self, api_name, api_version=None, permission_scope=None, force_refresh=False):
        """ Loads the requested service. If the service has already been loaded for this ServiceManager, return the previously loaded service (unless force_refresh is on). If it has not, delegate to get_credentials() as necessary.
        Args:
            api_name (str): The Google API name (i.e., 'drive.googleapis.com') or a shorthand version of it (e.g., 'drive').
            api_version (str, optional): The version of the API to use. If not provided, it will first check the DEFAULT_SCOPES mapping to determine if there is a default API. If not, it will attempt to use the directory service to determine the default API.
            permission_scope (str): The permission scope that accompanies this service. If the service_key is listed in the DEFAULT_SCOPES, this is optional and the default will be pulled. As fully-qualified permission scopes generally start with "https://www.googleapis.com/auth/", that part is optional and will be added if not present.
            force_refresh (bool): Reload the service even if it already has been loaded. Default is False.
        Returns:
            The loaded service.
        """
        #OLD and NEEDING TO DELETE: The fullly qualified service path, which also becomes the filename in the credential directory where the authorization token is stored. If not provided, this is generated from the service key ({service_key}.googleapis.com-python.json)

        if api_name in DEFAULT_SCOPES:
            keydata = DEFAULT_SCOPES[api_name]
            api_name = keydata[0]
            if not api_version:
                api_version = keydata[1]
            if not permission_scope:
                permission_scope = 'https://www.googleapis.com/auth/' + keydata[2]

        if not api_version:
            #TODO
            raise Exception("Pulling the discovery to determine the default version is not yet available")

        service_key = f"{api_name}.{api_version}"

        # Return the already-loaded service if it exists. This provides singleton-ish functionality at a Service:User level.
        if service_key in self.loaded_services and not force_refresh:
            return(self.loaded_services[service_key])

        # either service_key or service_path is required

        if self.api_key:
            # api keys should be used only for personal application and development
            # anyone with the api key can access Google services as the developer
            # TODO: This has not been tested in a while
            discovery = googleapiclient.discovery.build(api_name, api_version, developerKey=self.api_key)
        elif self.secret_file:
            # engage in OAuth2 authorization and credentialling

            service_path = f"{api_name}.{api_version}.googleapis.json"
            credentials = self.get_credentials(service_path, permission_scope)
            #authorizer = credentials.authorize(httplib2.Http())
            discovery = googleapiclient.discovery.build(api_name, api_version, credentials=credentials) #http=authorizer)
        else:
            raise Exception("Unable to load_service as no authentication models provided. Please specify an API Key or provide OAuth credentials to the googleapps.service.ServiceManager")

        # okay, save the service discovery for future calls
        self.loaded_services[service_key] = discovery
        return(self.loaded_services[service_key])


    def get_credentials(self, local_credential_file, permission_scope):
        """Gets valid user credentials - in most cases this is not required to be called directly and will be called from load_service().

        If nothing has been stored, or if the stored credentials are invalid, AND if application secret and scope are provided, the OAuth2 flow is completed to obtain the new credentials. This requires the user to authorize access via a web browser.

        Args:
            local_credential_file (path): the service-specific credential file, which is an authorization for this program/user to access these services. This file must exist in the credential_directory defined for this service object, or it will attempt to be fetched following user authorization. In the most common case where get_credentials is called by load_service(), this is determined in that function.
            permission_score (str): the permissions necessary for the application and that a remote call to authorize will request. The user may provide just the scope name (drive.readonly) or the full path to the permission scope (i.e. https://www.googleapis.com/auth/drive.readonly)

        Returns:
            Credentials, the obtained credential.
        """
        # auto create directory
        if not os.path.exists(self.credential_directory):
            os.makedirs(self.credential_directory)

        # convert the shorthand permission scope to the full path
        if not re.match('https', permission_scope):
            permission_scope = 'https://www.googleapis.com/auth/' + permission_scope


        # determine ifr the credentials have already been authorized and are active
        credential_path = os.path.join(self.credential_directory, local_credential_file)
        if os.path.exists(credential_path):
            credentials = google.oauth2.credentials.Credentials.from_authorized_user_file(credential_path, permission_scope)
        else:
            credentials = None

        if credentials and credentials.valid and credentials.expired and credentials.refresh_token:
            # creds exist and arre valid, but are expired - so just refresh
            credentials.refresh(google.auth.transport.requests.Request())
        else:
            # no credientials, or they are invalid!
            # -- get new credentials
            # identify the path to the credential's secret file.
            secret_path = os.path.join(self.credential_directory, self.secret_file)

            # a very detailed error message.
            if not os.path.exists(secret_path):
                raise FileNotFoundError(f"Could not find an OAuth secret file at {secret_path}, which is required for ServiceManager.get_credentials() to attempt to complete the authorization permission for scope {permission_scope} (the scope-level permission should be stored in {local_credential_file}, but it does not exist or is no longer valid and needs to be authenticated). \n\nYou will need to indicate the correct path to your OAuth Credential JSON secret file or download a new copy from the Google Developers Console located at https://console.developers.google.com/apis/credentials/ before trying again.")

            # All set authorize flow.
            flow = google_auth_oauthlib.flow.InstalledAppFlow.from_client_secrets_file(secret_path, permission_scope)
                credentials = flow.run_local_server(port=0)

        # Save the credentials for the next run
        with open(credential_path, 'w') as token:
            token.write(credentials.to_json())

        return(credentials)



import datetime
import re

import googleapps

import warnings
from pprint import pprint

#from .. import Spreadsheet, Format 

class Workbook(googleapps.sheets.Spreadsheet):

    # set up some aliases to existing functions
    get_worksheet_by_name = googleapps.sheets.Spreadsheet.get_sheet
    close = googleapps.sheets.Spreadsheet.commit
    
    @property
    def worksheets(self):
        return(self.sheets)
    
    @property
    def filename(self):
        return(self.title)
    
    @filename.setter
    def filename(self, new_filename):
        return(self.title(new_filename))
    
    def __init__(self, filename):
        # eventually should support all of the extra params
        # see http://xlsxwriter.readthedocs.io/workbook.html
        super().__init__()
        self._title = filename
        self.add_request('spreadsheets', 'create', 
            callback=self.process_json, 
            body={ 'properties': {'title': filename} },
        )        
        #self.commit()
      
    def _add_uncommited_sheet(self, **kwargs):
        new_sheet = googleapps.sheets.xlsxwriter.Worksheet(self, **kwargs)        
        self._sheets.append(new_sheet)
        return(new_sheet)
    
    def add_worksheet(self, name=None):
        #Add a new worksheet to a workbook.
        # In theory we could just alias to the superclass.....?
        
        if not len(self.sheets):
            # this is the first worksheet, which is automatically created
            # by google when we do the spreadshseets().create method. 
            # so, we just generate the uncommited reference and update
            # the sheet name.
            first_sheet = self._add_uncommited_sheet(title='Sheet1', id=0, index=0)
            if name:
                # this triggers a updateSheet request to set the title. 
                # TODO: we could potentially save this step if we went back and updated
                # the create() request to set the sheet name at that time
                first_sheet.title = name
                
            return(first_sheet)
        
        return(super().add_sheet(name))
    
    xlsx_horizontal_alignments = {'JUSTIFY':'CENTER', 'CENTER_ACROSS':'CENTER'} # < these don't exist in google        
    xlsx_vertical_alignments = {'VCENTER': 'MIDDLE', 
                              'VJUSTIFY':'MIDDLE'} # < this doesn't exist in google
        
        
    xlsxwriter_border_styles = {
        0: {'style':'NONE'},
        1: {'style':'SOLID'},
        2: {'style':'SOLID_MEDIUM'},
        3: {'style':'DASHED'},
        4: {'style':'DOTTED'},
        5: {'style':'SOLID_THICK'},
        6: {'style':'DOUBLE'},
        # the rest of these are best-example but cannot be exactly duplicated by the Google Sheets API
        7: {'style':'SOLID'}, #solid, 0
        8: {'style':'DASHED'}, #dashed, 2
        9: {'style':'DOTTED'}, #dash dot
        10: {'style':'DOTTED'}, #dash dot, 2
        11: {'style':'DASHED'}, # dash dot dot
        12: {'style':'DASHED'}, # dash dot dot, 2
        13: {'style':'DASHED'}, #slant dash dot
        }

    def add_format(self, format):
        #Create a new Format object to formats cells in worksheets.
        same_params = ('bold', 'text_wrap', 'font_size', 'italic', 'foreground') 
        mapped_params = {'bg_color': 'background'} 
        
        fmt = googleapps.sheets.Format()
        for formatkey in format:
            if formatkey in same_params:
                setattr(fmt, formatkey, format[formatkey])
        
            if formatkey in mapped_params:
                setattr(fmt, mapped_params[formatkey], format[formatkey])
        
        if 'num_format' in format:
            nf = format['num_format']
            if type(nf) is int:
                raise Exception("Have not figured this out yet")
            #import pdb
            #pdb.set_trace()
            if re.search('%', nf):
                fmt.number_format_type = 'PERCENT'
                fmt.number_format_pattern = nf
            else:
                fmt.number_format_type = 'NUMBER'
                fmt.number_format_pattern = re.sub('0', '#', nf)
#            else:
#                raise Exception("Cannot handle num format '" + str(nf) + "' yet")
                    
        if 'align' in format:
            # in xlswriter, align can handle both horizontal and vertical
            value = format['align'].upper()
            if value in googleapps.sheets.Format.horizontal_alignments:
                fmt.align = value
            elif value in googleapps.sheets.Format.vertical_alignments:
                fmt.vertical_align = value
            elif value in self.xlsx_horizontal_alignments:
                fmt.align = self.xlsx_horizontal_alignments[value]
            elif value in self.xlsx_vertical_alignments:
                fmt.vertical_align = self.xlsx_vertical_alignments[value]
            else:
                raise Exception("Unknown alignment " + format['align'].upper())

        if 'valign' in format:
            value = format['valign'].upper()
            if value in self.xlsx_vertical_alignments:
                fmt.vertical_align = self.xlsx_vertical_alignments[value]
            else:
                fmt.vertical_align = value
    
        if 'border' in format:
            fmt.border = True
    
        if 'locked' in format:
            fmt.locked = format['locked']
        else:
            fmt.locked = None
        return(fmt)
        
        
    @classmethod
    def translate_border_style(cls, style=1):
        #TODO
        return(cls.render_border_json(**xlsxwriter_border_styles[style]))
        xlsxwriter_border_styles = {
        0: {'style':'NONE'},
        1: {'style':'SOLID'},
        2: {'style':'SOLID_MEDIUM'},
        3: {'style':'DASHED'},
        4: {'style':'DOTTED'},
        5: {'style':'SOLID_THICK'},
        6: {'style':'DOUBLE'},
        # the rest of these are best-example but cannot be exactly duplicated by the Google Sheets API
        7: {'style':'SOLID'}, #solid, 0
        8: {'style':'DASHED'}, #dashed, 2
        9: {'style':'DOTTED'}, #dash dot
        10: {'style':'DOTTED'}, #dash dot, 2
        11: {'style':'DASHED'}, # dash dot dot
        12: {'style':'DASHED'}, # dash dot dot, 2
        13: {'style':'DASHED'}, #slant dash dot
        }
    
           
                
                
                
                
                

   


# Copyright (c) 2019-2021 Kevin Crouse
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# @license: http://www.apache.org/licenses/LICENSE-2.0
# @author: Kevin Crouse (krcrouse@gmail.com)

import os, os.path
import datetime
import pickle
import re
import shutil

import warnings
from pprint import pprint

import googleapps
from googleapps.serviceclient import CommitType
from .driveitem import DriveItem
from .file import File

class Folder (DriveItem):

    def __init__(self, *args, **kwargs):
        super().__init__(*args,**kwargs)
        self._children = None

    @property
    def mime_type(self):
        return(googleapps.drive.MimeType.folder_mime_type())

    def process_request_json(self, json):
        if 'mimeType' in json and json['mimeType'] != self.mime_type:
            raise Exception("Processing Google Drive remote response for (alleged) folder " + str(json['id']) + ", Drive item is not reported as a folder but as type '" + json['mimeType']+ "'")
        del json['mimeType']
        return(super().process_request_json(json))

    @property
    def children(self):
        if self._children is None:
            self._children = googleapps.drive.find_all(in_folder=self, all_drives=self.in_shared_drive)
        return(self._children)

    def get_children(self, name=None, use_regexp=False, mime_type=None):
        files = []
        for child in self.children:
            if mime_type and child.mime_type != mime_type:
                continue
            if name:
                if use_regexp:
                    if re.search(name, child.name, re.I):
                        files.append(child)
                elif name == child.name:
                    files.append(child)
            else:
                files.append(child)
        return(files)

    def get_folder(self, *args, **kwargs):
        folders = self.get_folders(*args, **kwargs)
        if len(folders) == 0:
            return
        if len(folders) == 1:
            return(folders[0])

        raise Exception("Multiple folders match the following criteria: " + str(kwargs))

    def get_file(self, *args, **kwargs):
        files = self.get_files(*args, **kwargs)
        if len(files) == 0:
            return
        if len(files) == 1:
            return(files[0])

        raise Exception("Multiple files match criteria: " + str(kwargs))


    def get_files(self, *args, **kwargs):
        children= self.get_children(*args, **kwargs)
        files = []
        folder_mimetype = googleapps.drive.MimeType.folder_mime_type()
        for child in children:
            if child.mime_type == folder_mimetype:
                continue
            files.append(child)
        return(files)

    def get_folders(self, *args, **kwargs):
        return(self.get_children(*args, mime_type=googleapps.drive.MimeType.folder_mime_type(), **kwargs))


    def create_folder(self, name=None, **body):
        if 'mime_type' in body:
            raise Exception("Parameter mime_type not valid for create_folder()")

        if name:
            body['name'] = name

        body['parents'] = [self.id]

        item = Folder(**body)
        # only save if autocommit is on
        if item.commit_policy >= CommitType.auto_commit:
            item.save()
        # add it
        self.children.append(item)

        return(item)


    def upload_file(self, filepath, mime_type=None, convert=None, **parameters):
        if 'parents' in parameters:
            parameters['parents'].append(self.id)
        else:
            parameters['parents'] = [self.id]
        obj = File(**parameters)
        obj.upload(filepath, mime_type=mime_type, convert=convert)
        self.children.append(obj)
        return(obj)

    def upload(self, folderpath):
        # create the folder
        raise Exception ("Upload folder not implemented")

        # iterate
        #for root, dirs, files in os.scandir(folderpath):

            #for file in files:


    #-------- OLD -------------
    @classmethod
    def remove_by_id(self, id):
        raise Exception("Not implemented yet")

    @classmethod
    def remove_by_id(self, id):
        raise Exception("Not implemented yet")

    def make_path(self, path):
        raise Exception("Not implemented yet")

    def remove(self, if_empty=False):
        raise Exception("Not implemented yet")
#    def rmtree(self):
#        raise Exception("Not implemented yet")

    def copy(self):
        raise Exception("Not implemented yet")

    def download(self):
        raise Exception("Not implemented yet")

    def make_archive(self, output_name, format='zip'):
        # formats: tar gztar bztar xztar
        raise Exception("Not implemented yet")

    def scan(self, files_only=None, folders_only=None):
        if files_only:
            is_folder = False
        elif folders_only:
            is_folder = True
        else:
            is_folder = None
        return(googleapps.drive.find(self.service, in_folder=self.id, is_folder=is_folder, all_drives=self.in_shared_drive,))

    def list(self, *args, result_type='name', **kwargs):
        # just shorthand to call scanfolder
        # returns a list of folder names
        folders = self.scanfolder(*args, **kwargs)
        result = []
        for item in folders:
            if result_type == 'name':
                result.append(item.name)
            elif result_type == 'id':
                result.append(item.id)
            else:
                raise Exception("google.drive.listfolder called with unknown result type " + str(result_type))
        return(result)

    def walk(self):
        raise Exception("Not thoroughly investigated.")

        q = "'" + str(self.id) + "' in parents"
        if not include_trash:
            q += " and trashed=false"

        if filters:
            if type(filters) is not list:
                filters = [ filters ]

            for ftr in filters:
                if ftr == 'folders':
                    q += " and mimeType='application/vnd.google-apps.folder'"
                elif ftr == 'files':
                    q += " and mimeType='application/vnd.google-apps.folder'"
                elif ftr == 'spreadsheets':
                    q += " and mimeType='application/vnd.google-apps.spreadsheet'"
                else:
                    raise Exception("Unknown filter requested, type" + str(ftr))

        result = []
        for item in self.default_service.files().list(
                    q=q,
                    fields="nextPageToken, files(id, name, parents, trashed)",
            ).execute().get('files'):

            result.append(DriveItem(item['id'], name=item['name']))

        return(result)
